"""
Copy a KGTK file, validating it and producing a clean KGTK file (no
comments, whitespace lines, etc.) as output.

TODO: Need KgtkWriterOptions.

TODO: Need to plumn the infrastructure so we can report at least
a count of how many repair actions took place (per action type).
Ideally, we'ld like the optino to log individual repair actions.

TODO: Add a reject file.

"""

from argparse import Namespace, SUPPRESS

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Validate a KGTK file and output a clean copy: no comments, whitespace lines, invalid lines, etc. ',
        'description': 'Validate a KGTK file and output a clean copy. ' +
        'Empty lines, whitespace lines, comment lines, and lines with empty required fields are silently skipped. ' +
        'Header errors cause an immediate exception. Data value errors are reported and the line containing them skipped. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert clean-data --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions
    
    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True)
    parser.add_output_file(positional=True)
    parser.add_output_file(who="Reject file",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    validate_by_default=True,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = False,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        **kwargs # Whatever KgtkReaderOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    import typing
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file_path: Path = KGTKArgumentParser.get_output_file(output_file)
    reject_kgtk_file_path: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="Reject file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file_path), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file_path), file=error_file)
        if reject_kgtk_file_path is not None:
            print("--reject-file=%s" % str(reject_kgtk_file_path), file=error_file)
            
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    if verbose:
        if str(input_kgtk_file_path) == "-":
            print ("Cleaning data from stdin", file=error_file, flush=True)
        else:
            print("Cleaning data from '%s'" % str(input_kgtk_file_path), file=error_file, flush=True)
        if str(output_kgtk_file_path) == "-":
            print ("Writing data to stdout", file=error_file, flush=True)
        else:
            print("Writing data to '%s'" % str(output_kgtk_file_path), file=error_file, flush=True)
        if str(reject_kgtk_file_path) == "-":
            print ("Writing reject data to stdout", file=error_file, flush=True)
        else:
            print("Writing reject data to '%s'" % str(reject_kgtk_file_path), file=error_file, flush=True)

    reject_kgtk_file: typing.Optional[typing.TextIO] = None
    if reject_kgtk_file_path is not None:
        reject_kgtk_file = open(reject_kgtk_file_path, mode="wt")

    try:
        kr: KgtkReader = KgtkReader.open(input_kgtk_file_path,
                                         error_file=error_file,
                                         reject_file=reject_kgtk_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         output_kgtk_file_path,
                                         verbose=verbose, very_verbose=very_verbose)
        
        line_count: int = 0
        row: typing.List[str]
        for row in kr:
            kw.write(row)
            line_count += 1

        kw.close()
        if reject_kgtk_file is not None:
            reject_kgtk_file.close()

        if verbose:
            print("Copied %d clean data lines" % line_count, file=error_file, flush=True)
        return 0

    except Exception as e:
        raise KGTKException(e)




"""
KGTK core SQL functions
"""

import sys
import re
from   functools import lru_cache

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *
from   kgtk.kypher.functions import *
import kgtk.kypher.parser as parser


### Core functions:

# Naming convention: a suffix of _string indicates that the resulting
# value will be additionally converted to a KGTK string literal.  The
# same could generally be achieved by calling 'kgtk_stringify' explicitly.

# Regular expressions:

@lru_cache(maxsize=100)
def _get_regex(regex):
    return re.compile(regex)
    
@sqlfun(num_params=2, deterministic=True)
def kgtk_regex(x, regex):
    """Regex matcher that implements the Cypher '=~' semantics which must match the whole string.
    """
    # TO DO: using the new API, we could create query-translation-time compilations of the regex
    m = isinstance(x, str) and _get_regex(regex).match(x) or None
    return m is not None and m.end() == len(x)


# NULL value utilities:

# In the KGTK file format we cannot distinguish between empty and NULL values.
# Both KGTKReader and SQLite map missing values onto empty strings, however,
# database functions as well as our KGTK user functions return NULL for undefined
# values.  These can be tested via 'IS [NOT] NULL', however, in some cases it is
# convenient to convert from one to the other for more uniform tests and queries.

@sqlfun(num_params=1, deterministic=True)
def kgtk_null_to_empty(x):
    """If 'x' is NULL map it onto the empty string, otherwise return 'x' unmodified.
    """
    if x is None:
        return ''
    else:
        return x

@sqlfun(num_params=1, deterministic=True)
def kgtk_empty_to_null(x):
    """If 'x' is the empty string, map it onto NULL, otherwise return 'x' unmodified.
    """
    if x == '':
        return None
    else:
        return x


# Python eval:

_sqlstore_module = sys.modules['kgtk.kypher.sqlstore']
_builtins_module = sys.modules['builtins']

def get_pyeval_fn(fnname):
    pos = fnname.rfind('.')
    if pos < 0:
        return getattr(_sqlstore_module, fnname, None) or getattr(_builtins_module, fnname)
    else:
        # we lookup the module name relative to this module in case somebody imported an alias:
        return getattr(getattr(_sqlstore_module, fnname[0:pos]), fnname[pos+1:])

@sqlfun(num_params=-1)
def pyeval(*expression):
    """Python-eval 'expression' and return the result (coerce value to string if necessary).
    Multiple 'expression' arguments will be concatenated first.
    """
    try:
        val = eval(''.join(expression))
        return isinstance(val, (str, int, float)) and val or str(val)
    except:
        pass

@sqlfun(num_params=-1)
def pycall(fun, *arg):
    """Python-call 'fun(arg...)' and return the result (coerce value to string if necessary).
    'fun' must name a function and may be qualified with a module imported by --import.
    """
    try:
        val = get_pyeval_fn(fun)(*arg)
        return isinstance(val, (str, int, float)) and val or str(val)
    except:
        pass


# Aggregate functions

# support SqlFunction.is_aggregate() test:
BuiltinAggregateFunction(name='avg').define()
BuiltinAggregateFunction(name='count').define()
BuiltinAggregateFunction(name='group_concat').define()
BuiltinAggregateFunction(name='max').define()
BuiltinAggregateFunction(name='min').define()
BuiltinAggregateFunction(name='sum').define()
BuiltinAggregateFunction(name='total').define()


# Special SQLite and Kypher functions:

class Cast(BuiltinFunction):
    """Implement SQLite CAST as a Kypher function (no equivalent in Cypher).
    """

    def translate_call_to_sql(self, query, expr, state):
        """Translate CAST(expr, type) to SQL.
        """
        if len(expr.args) == 2 and isinstance(expr.args[1], parser.Variable):
            arg = query.expression_to_sql(expr.args[0], state)
            typ = expr.args[1].name
            return f'{self.get_name()}({arg} AS {typ})'
        else:
            raise KGTKException("Illegal CAST expression")

Cast(name='cast').define()

class Likelihood(BuiltinFunction):
    """Special-case SQLite LIKELIHOOD translation.
    """

    def translate_call_to_sql(self, query, expr, state):
        """Translate SQLite LIKELIHOOD(expr, prob) expression to SQL which needs a compile-time constant
        for its probability argument instead of the value parameters we usually use to supply literals.
        """
        if len(expr.args) == 2 and isinstance(expr.args[1], parser.Literal) and isinstance(expr.args[1].value, (int, float)):
            arg = query.expression_to_sql(expr.args[0], state)
            prob = expr.args[1].value
            return f'{self.get_name()}({arg}, {prob})'
        else:
            raise Exception("Illegal LIKELIHOOD expression")

Likelihood(name='likelihood').define()

class Concat(BuiltinFunction):
    """Special-case Cypher's CONCAT function which is handled by SQLite's ||-operator.
    """
    
    def translate_call_to_sql(self, query, expr, state):
        """Translate CONCAT(arg...) to SQL using SQLite's ||-operator.
        """
        args = [query.expression_to_sql(arg, state) for arg in expr.args]
        return f'({" || ".join(args)})'

Concat(name='concat').define()

class RowId(BuiltinFunction):
    """Implement table.rowid lookup for Kypher query variables.
    """

    def translate_call_to_sql(self, query, expr, state):
        """Translate ROWID(var) to SQL.  rowid's are 1-based and are an efficient
        way to count the number of rows in a table, for example, via 'max(rowid(r))'.
        If the 'var' argument is a join variable, it is ambiguous which graph table
        it should be associated with, in which case the choice will be random and
        might lead to unexpected results.
        """
        if len(expr.args) == 1 and isinstance(expr.args[0], parser.Variable):
            graph, _, _ = query.variable_to_sql(expr.args[0], state)
            return f'{graph}.rowid'
        else:
            raise KGTKException("Illegal ROWID expression")

RowId(name='rowid').define()


### Experimental transitive taxonomy relation indexing:

@lru_cache(maxsize=1000)
def kgtk_decode_taxonomy_node_intervals(intervals):
    """Decode a difference-encoded list of 'intervals' into a numpy array with full intervals.
    """
    # expensive imports we don't want to run unless needed, lru cache will eliminate repeat overhead:
    import gzip, binascii, numpy
    if intervals[0] == 'z':
        intervals = gzip.decompress(binascii.a2b_base64(intervals[1:])).decode()
    intervals = intervals.replace(';', ',0,')
    if intervals.endswith(','):
        intervals = intervals[0:-1]
    intervals = list(map(int, intervals.split(',')))
    # we special-case single intervals and binary search on more than one interval:
    if len(intervals) > 2:
        # add sentinel, so we always have a sort insertion point before the end of the array:
        intervals.append(0)
    intervals = numpy.array(intervals, dtype=numpy.int32)
    # decode difference encoding:
    for i in range(1, len(intervals)):
        intervals[i] += intervals[i-1]
    if len(intervals) > 2:
        # initialize sentinel:
        intervals[-1] = 2**31 - 1
    return intervals

# timing on 2.5M calls:
# - just call and return: Q123: 0.95s, Q5: 1.05s
# - int(label):           Q123: 1.38s, Q5: 1.45s
# - decode intervals:     Q123: 1.68s, Q5: 2.12s
# - single int range:     Q123: 3.10s, Q5: 2.20s
# - single int >=,<=:     Q123: 2.60s, Q5: 2.20s
# - range shortcut:       Q123: 2.60s, Q5: 2.20s
# - searchsorted:         Q123: 2.60s, Q5: 4.90s
# - result1:              Q123: 2.60s, Q5:11.10s
# - result2: (wrong)      Q123: 2.60s, Q5: 6.95s
# - result3:              Q123: 2.60s, Q5:10.80s 
# - result4: (wrong)      Q123: 2.60s, Q5: 6.30s
# - result5:              Q123: 2.60s, Q5: 6.60s
# - bool(result5)         Q123: 2.60s, Q5: 6.70s

@sqlfun(num_params=2, deterministic=True)
def kgtk_is_subnode(label, encoded_intervals):
    """Return True if 'label' is contained in one of the encoded 'intervals'.
    'intervals' is a flat list of sorted, closed integer intervals.
    """
    # NOTE: it took us a while to optimize this properly; the crucial bit was
    # to use 'int' to cast array elements before comparing them via >=,<= and ==
    label = int(label)
    # cached lookup is fast, trying to use a shorter key string (e.g., edge ID) does not help:
    intervals = kgtk_decode_taxonomy_node_intervals(encoded_intervals)
    # check single interval shortcut:
    if len(intervals) == 2:
        # "casting" to int first significantly speeds things up (also beats 'range'):
        return label >= int(intervals[0]) and label <= int(intervals[1])
    i = intervals.searchsorted(label)
    # this runs on lists but is 3x slower, not sure why, it says there is a C-implementation:
    #i = bisect.bisect_left(intervals, label)
    #result1 = (i & 1) or (i < len(intervals) and intervals[i] == label)
    #result2 = (i & 1) or (i < len(intervals) and intervals[i] is label)
    #result3 = (i & 1) or (intervals[i] == label)
    #result4 = (i & 1) or (intervals[i] is label)
    # "casting" to int first gives us a much faster equality test:
    result5 = (i & 1) or (int(intervals[i]) == label)
    #sys.stderr.write('%s  %s  %s\n' % (label, intervals, result))
    # TO DO: figure out whether we should add this to all predicates above:
    return bool(result5)



"""
Join two KGTK edge files or two KGTK node files.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Join two KGTK files',
        'description': """Join two KGTK edge files or two KGTK node files.

Join keys are extracted from one or both input files and stored in memory,
then the data files are processed in a second pass.  stdin will not work as an
input file if join keys are needed from it.

The output file contains the union of the columns in the two
input files, adjusted for predefined name aliasing.

Specify --left-join to get a left outer join.
        The output file will contain all records from the
        left input file, along with records from the right
        input file with matching join column values.

Specify --right-join to get a right outer join.
        The output file will contain all records from the
        right input file, along with records from the left
        input file with matching join column values.

Specify both --left-join and --right-join to get a full outer
join (equivalent to cat or set union).
        The output file will contain all records from both
        the left input file and the right input file.

Specify neither --left-join nor --right-join to get an inner
join.  If there are no columns beyond the join columns, then
this is equivalent to set intersection.
        The output file will contain records from the left
        input file and from the right input file for which
        the join column value match.

By default, node files are joined on the id column, while edge files are joined
on the node1 column. The label and node2 columns may be added to the edge file
join criteria.  Alternatively, the left and right file join columns may be
listed explicitly.

To join an edge file to a node file, or to join quasi-KGTK files, use the
following option (enable expert mode for more information):

--mode=NONE

Expert mode provides additional command arguments.
"""
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.join.kgtkjoiner import KgtkJoiner
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="The left-side KGTK file to join (required).",
                          dest="left_file",
                          options=["--left-file"],
                          metavar="LEFT_FILE",
                          positional=True)

    parser.add_input_file(who="The right-side KGTK file to join (required).",
                          dest="right_file",
                          options=["--right-file"],
                          metavar="RIGHT_FILE",
                          positional=True)

    parser.add_output_file()

    parser.add_argument(      "--join-on-id", dest="join_on_id",
                              help="If both input files are edge files, include the id column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--join-on-label", dest="join_on_label",
                              help="If both input files are edge files, include the label column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--join-on-node2", dest="join_on_node2",
                              help="If both input files are edge files, include the node2 column in the join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)
    
    parser.add_argument(      "--left-prefix", dest="left_prefix",
                              help="An optional prefix applied to left file column names in the output file (default=None).")
    
    parser.add_argument(      "--left-file-join-columns", dest="left_join_columns", help="Left file join columns (default=None).", nargs='+')

    parser.add_argument(      "--left-join", dest="left_join", help="Perform a left outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--right-prefix", "--prefix", dest="right_prefix",
                              help="An optional prefix applied to right file column names in the output file (default=None).")
    
    parser.add_argument(      "--right-file-join-columns", dest="right_join_columns", help="Right file join columns (default=None).", nargs='+')
    
    parser.add_argument(      "--right-join", dest="right_join", help="Perform a right outer join (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--field-separator", dest="field_separator",
                              help=h("Separator for multifield keys (default=%(default)s)")
                              , default=KgtkJoiner.FIELD_SEPARATOR_DEFAULT)

    # Build the command arguments. File arguments can be set for individual
    # files, or for all files.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="left", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="right", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(left_file: KGTKFiles,
        right_file: KGTKFiles,
        output_file: KGTKFiles,

        join_on_id: bool = False,
        join_on_label: bool = False,
        join_on_node2: bool = False,

        left_prefix: typing.Optional[str] = None,
        left_join_columns: typing.Optional[typing.List[str]] = None,
        left_join: bool = False,

        right_prefix: typing.Optional[str] = None,
        right_join_columns: typing.Optional[typing.List[str]] = None,
        right_join: bool = False,

        field_separator: typing.Optional[str] = None,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.kgtkjoiner import KgtkJoiner
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    left_file_path: Path = KGTKArgumentParser.get_input_file(left_file, who="KGTK left file")
    right_file_path: Path = KGTKArgumentParser.get_input_file(right_file, who="KGTK right file")
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)

    field_separator = KgtkJoiner.FIELD_SEPARATOR_DEFAULT if field_separator is None else field_separator

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    if not right_join:
        if str(left_file_path) == "-":
            print("The left file may not be stdin when an inner join or left join is requested.", file=error_file, flush=True)
            return 1

    if not left_join:
        if str(right_file_path) == "-":
            print("The right file may not be stdin when an inner join or right join is requested.", file=error_file, flush=True)
            return 1

    if str(left_file_path) == "-" and str(right_file_path) == "-":
        print("The left and right files may not both be stdin.", file=error_file, flush=True)
        return 1

    # Build the option structures.
    left_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="left", fallback=True)
    right_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="right", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        # TODO: left_file_path, right_file_path, --join-on-label, etc.
        print("--left-file=%s" % str(left_file_path), file=error_file)
        print("--right-file=%s" % str(right_file_path), file=error_file)
        print("--output-file=%s" % str(output_file_path), file=error_file)
        
        print("--left-join=%s" % str(left_join), file=error_file)
        print("--right-join=%s" % str(right_join), file=error_file)
        print("--join-on-id=%s" % str(join_on_id), file=error_file)
        print("--join-on-label=%s" % str(join_on_label), file=error_file)
        print("--join-on-node2=%s" % str(join_on_node2), file=error_file)
        if left_join_columns is not None:
            print("--left-join-columns=%s" % " ".join(left_join_columns), file=error_file)
        if right_join_columns is not None:
            print("--right-join-columns=%s" % " ".join(right_join_columns), file=error_file)
        if left_prefix is not None:
            print("--left-prefix=%s" % str(left_prefix), file=error_file)
        if right_prefix is not None:
            print("--right-prefix=%s" % str(right_prefix), file=error_file)
        print("--field-separator=%s" % repr(field_separator), file=error_file)
              
        left_reader_options.show(out=error_file, who="left")
        right_reader_options.show(out=error_file, who="right")
        value_options.show(out=error_file)

    try:
        kr: KgtkJoiner = KgtkJoiner(
            left_file_path=left_file_path,
            right_file_path=right_file_path,
            output_path=output_file_path,
            left_join=left_join,
            right_join=right_join,
            join_on_id=join_on_id,
            join_on_label=join_on_label,
            join_on_node2=join_on_node2,
            left_join_columns=left_join_columns,
            right_join_columns=right_join_columns,
            left_prefix=left_prefix,
            right_prefix=right_prefix,
            field_separator=field_separator,
            left_reader_options=left_reader_options,
            right_reader_options=right_reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        kr.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
SQLStore to support Kypher queries over KGTK graphs.
"""

from   functools import lru_cache
import copy

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *


class KgtkInfoTable(object):
    """API for access to file and graph info using a fine-grained KGTK edge data model.
    """
    
    TABLE_SCHEMA = sdict[
        '_name_': '<NAME>',
        'columns': sdict[
            'node1':  sdict['_name_': 'node1',  'type': 'TEXT',],
            'label':  sdict['_name_': 'label',  'type': 'TEXT',],
            'node2':  sdict['_name_': 'node2',  'type': 'TEXT',],
            'id':     sdict['_name_': 'id',     'type': 'TEXT',],
        ],
    ]

    def __init__(self, store, table):
        """Create an info table object for 'table' stored in 'store'.
        """
        self.store = store
        self.table = table
        self.schema = copy.deepcopy(self.TABLE_SCHEMA)
        self.schema._name_ = self.table
        self.columns = self.schema.columns
        self.column_list = self.store.get_full_column_list(self.schema)
        self.init_table()

    def init_table(self):
        """If the info table doesn't exist yet, define it from its schema.
        """
        if not self.store.has_table(self.table):
            self.store.execute(self.store.get_table_definition(self.schema))
            table = self.table
            node1_idx = f'{table}_node1_idx'
            stmt = f'CREATE INDEX {sql_quote_ident(node1_idx)} ON {sql_quote_ident(table)} ({self.columns.node1._name_})'
            self.store.execute(stmt)
            label_idx = f'{table}_label_idx'
            stmt = f'CREATE INDEX {sql_quote_ident(label_idx)} ON {sql_quote_ident(table)} ({self.columns.label._name_})'
            self.store.execute(stmt)
            node2_idx = f'{table}_node2_idx'
            stmt = f'CREATE INDEX {sql_quote_ident(node2_idx)} ON {sql_quote_ident(table)} ({self.columns.node2._name_})'
            self.store.execute(stmt)

    def make_object_id(self, prefix, *keys):
        """Create a unique object ID starting with 'prefix' and based on 'keys'.
        """
        # we only use this during DB updates, so the import overhead should be negligible:
        import shortuuid
        return prefix + shortuuid.uuid(r'_/^^\_'.join(map(str, keys)))

    @lru_cache(maxsize=None)
    def get_edge_pattern_restriction(self, node1=None, label=None, node2=None, id=None):
        # this should be called with a True/False binding pattern
        # instead of actual values for maximum caching effectiveness:
        restrictions = []
        if node1:
            restrictions.append(f'{sql_quote_ident(self.columns.node1._name_)}=?')
        if label:
            restrictions.append(f'{sql_quote_ident(self.columns.label._name_)}=?')
        if node2:
            restrictions.append(f'{sql_quote_ident(self.columns.node2._name_)}=?')
        if id:
            restrictions.append(f'{sql_quote_ident(self.columns.id._name_)}=?')
        return ' AND '.join(restrictions)

    @lru_cache(maxsize=None)
    def get_edges_query(self, node1=None, label=None, node2=None, id=None):
        # this should be called with a True/False binding pattern
        # instead of actual values for maximum caching effectiveness:
        table = sql_quote_ident(self.table)
        return_columns = self.column_list
        restrictions = self.get_edge_pattern_restriction(node1, label, node2, id) or "TRUE"
        query = f'SELECT {return_columns} FROM {table} WHERE {restrictions}'
        return query        
        
    def get_edges(self, node1=None, label=None, node2=None, id=None):
        """Return an iterator generating the edges identified by the restriction pattern
        defined through the values of 'node1/label/node2/id'.
        """
        pattern = [node1, label, node2, id]
        query = self.get_edges_query(*[x is not None for x in pattern])
        return self.store.execute(query, [x for x in pattern if x is not None])

    def add_edge(self, node1, label, node2, id=None):
        """Add an edge defined by 'node1/label/node2'.  If no 'id' is provided,
        generate one based purely on the triple.  This does not check for existing
        duplicate edges and will simply append a new row to the table.
        """
        if id is None:
            id = self.make_object_id('e-', node1, label, node2)
        table = sql_quote_ident(self.table)
        columns = self.column_list
        stmt = f'INSERT INTO {table} ({columns}) VALUES (?,?,?,?)'
        self.store.execute(stmt, (node1, label, node2, id))

    def delete_edges(self, node1=None, label=None, node2=None, id=None):
        """Delete all edges identified by the restriction pattern
        defined through the values of 'node1/label/node2/id'.
        """
        pattern = [node1, label, node2, id]
        table = sql_quote_ident(self.table)
        restrictions = self.get_edge_pattern_restriction(*[x is not None for x in pattern]) or "TRUE"
        stmt = f'DELETE FROM {table} WHERE {restrictions}'
        self.store.execute(stmt, [x for x in pattern if x is not None])

    def update_edge(self, node1, label, node2, id=None):
        """Update all edges identified by 'node1/label' to have 'node2' and 'id'
        as their remaining elements.  An id will be generated if necessary.
        """
        if id is None:
            id = self.make_object_id('e-', node1, label, node2)
        table = sql_quote_ident(self.table)
        columns = self.get_edge_pattern_restriction(False, False, True, True)
        columns = columns.replace('=? AND ', '=?, ')
        restrictions = self.get_edge_pattern_restriction(True, True, False, False)
        stmt = f'UPDATE {table} SET {columns} WHERE {restrictions}'
        self.store.execute(stmt, (node2, id, node1, label))

    def get_values(self, node1, label):
        """Return the list of all values for 'node1/label'.
        """
        return [node2 for _, _, node2, _ in self.get_edges(node1, label)]

    def get_value(self, node1, label):
        """Return the first value for 'node1/label' or None if no such edge exists.
        """      
        for _, _, node2, _ in self.get_edges(node1, label):
            return node2
        return None
    
    def get_objects(self, node2, label):
        """Inverse to 'get_values' starting from 'node2/label' to objects 'node1'.
        """
        return [node1 for node1, _, _, _ in self.get_edges(None, label, node2)]

    def get_object(self, node2, label):
        """Inverse to 'get_value' starting from 'node2/label' to object 'node1'.
        Return None if no such edge exists.
        """
        for node1, _, _, _ in self.get_edges(None, label, node2):
            return node1
        return None

    def set_value(self, node1, label, node2):
        """Set the value of 'node1/label' to 'node2'  This deletes all preexisting
        matching edges and ensures only a single such edge exists after the operation.
        """
        self.delete_edges(node1, label)
        self.add_edge(node1, label, node2)

    def add_value(self, node1, label, node2):
        """Add a single 'node1/label/node2' edge and remove any duplicates.
        """
        self.delete_edges(node1, label, node2)
        self.add_edge(node1, label, node2)
    
    def update_value(self, node1, label, node2):
        """Change the node2 value of a 'node1/label' edge to 'node2'.
        Expects that exactly one such edge exists before the operation.
        """
        values = self.get_values(node1, label)
        if len(values) != 1:
            raise KGTKException('can only update existing single-valued edge')
        self.update_edge(node1, label, node2)
        
    def get_all_objects(self):
        """Return a list of distinct 'node1' objects in this table.
        """
        query = f'SELECT DISTINCT {sql_quote_ident(self.columns.node1._name_)} FROM {sql_quote_ident(self.table)}'
        return [node1 for (node1,) in self.store.execute(query)]



"""Import an ntriples file, writing a KGTK file.

TODO: Need KgtkWriterOptions
"""

from argparse import _MutuallyExclusiveGroup, Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import an ntriples file.',
        'description': 'Import an ntriples file, writing a KGTK file.' + 
        '\n\nAdditional options are shown in expert help.\nkgtk --expert import-ntriples --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.imports.kgtkntriples import KgtkNtriples
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="The ntriples file(s) to import.", allow_list=True)
    parser.add_output_file()

    parser.add_output_file(who="The ntriples output file for records that are rejected.",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    parser.add_input_file(who="The KGTK input file with known namespaces.",
                          dest="namespace_file",
                          options=["--namespace-file"],
                          metavar="NAMESPACE_FILE",
                          optional=True)

    parser.add_output_file(who="The KGTK output file with updated namespaces.",
                           dest="updated_namespace_file",
                           options=["--updated-namespace-file"],
                           metavar="NAMESPACE_FILE",
                           optional=True)

    KgtkNtriples.add_arguments(parser)
    KgtkIdBuilderOptions.add_arguments(parser)
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,

        namespace_file: KGTKFiles,
        updated_namespace_file: KGTKFiles,

        namespace_id_prefix: str,
        namespace_id_use_uuid: bool,
        namespace_id_counter: int,
        namespace_id_zfill: int,
        output_only_used_namespaces: bool,
        build_new_namespaces: bool,

        allow_lax_uri: bool,
        allow_unknown_datatype_iris: bool,
        allow_turtle_quotes: bool,
        allow_lang_string_datatype: bool,
        lang_string_tag: str,

        local_namespace_prefix: str,
        local_namespace_use_uuid: bool,

        prefix_expansion_label: str,
        structured_value_label: str,
        structured_uri_label: str,

        newnode_prefix: str,
        newnode_use_uuid: bool,
        newnode_counter: int,
        newnode_zfill: int,

        build_id: bool,

        build_datatype_column: bool,
        datatype_column_name: str,

        validate: bool,
        summary: bool,

        override_uuid: typing.Optional[str],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.imports.kgtkntriples import KgtkNtriples
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    # Select where to send error messages, defaulting to stderr.
    input_file_paths: typing.List[Path] = KGTKArgumentParser.get_input_file_list(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    reject_file_path: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="KGTK reject file")

    namespace_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_input_file(namespace_file, who="KGTK namespace file")
    updated_namespace_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(updated_namespace_file, who="KGTK updated namespace file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)    
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-files %s" % " ".join([repr(str(path)) for path in input_file_paths]), file=error_file, flush=True)
        print("--output-file=%s" % repr(str(output_kgtk_file)), file=error_file, flush=True)
        if reject_file_path is not None:
            print("--reject-file=%s" % repr(str(reject_file_path)), file=error_file, flush=True)
        if namespace_kgtk_file is not None:
            print("--namespace-file=%s" % repr(str(namespace_kgtk_file)), file=error_file, flush=True)
        if updated_namespace_kgtk_file is not None:
            print("--updated-namespace-file=%s" % repr(str(updated_namespace_kgtk_file)), file=error_file, flush=True)

        print("--namespace-id-prefix %s" % repr(namespace_id_prefix), file=error_file, flush=True)
        print("--namespace-id-use-uuid %s" % repr(namespace_id_use_uuid), file=error_file, flush=True)
        print("--namespace-id-counter %s" % repr(namespace_id_counter), file=error_file, flush=True)
        print("--namespace-id-zfill %s" % repr(namespace_id_zfill), file=error_file, flush=True)
        print("--output-only-used-namespaces %s" % repr(output_only_used_namespaces), file=error_file, flush=True)
        print("--build-new-namespaces %s" % repr(build_new_namespaces), file=error_file, flush=True)

        print("--allow-lax-uri %s" % repr(allow_lax_uri), file=error_file, flush=True)
        print("--allow-unknown-datatype-iris %s" % repr(allow_unknown_datatype_iris), file=error_file, flush=True)
        print("--allow-turtle-quotes %s" % repr(allow_turtle_quotes), file=error_file, flush=True)
        print("--allow-lang-string-datatype %s" % repr(allow_lang_string_datatype), file=error_file, flush=True)
        print("--lang-string-tag %s" % repr(lang_string_tag), file=error_file, flush=True)
        
        print("--local-namespace-prefix %s" % repr(local_namespace_prefix), file=error_file, flush=True)
        print("--local-namespace-use-uuid %s" % repr(local_namespace_use_uuid), file=error_file, flush=True)

        print("--prefix-expansion-label %s" % repr(prefix_expansion_label), file=error_file, flush=True)
        print("--structured-value-label %s" % repr(structured_value_label), file=error_file, flush=True)
        print("--structured-uri-label %s" % repr(structured_uri_label), file=error_file, flush=True)
        
        print("--newnode-prefix %s" % repr(newnode_prefix), file=error_file, flush=True)
        print("--newnode-use-uuid %s" % repr(newnode_use_uuid), file=error_file, flush=True)
        print("--newnode-counter %s" % repr(newnode_counter), file=error_file, flush=True)
        print("--newnode-zfill %s" % repr(newnode_zfill), file=error_file, flush=True)
        
        print("--build-id=%s" % repr(build_id), file=error_file, flush=True)

        print("--build-datatype-column %s" % repr(build_datatype_column), file=error_file, flush=True)
        print("--datatype-column-name %s" % repr(datatype_column_name), file=error_file, flush=True)

        print("--validate=%s" % repr(validate), file=error_file, flush=True)
        print("--summary=%s" % repr(summary), file=error_file, flush=True)
        
        print("--override-uuid=%s" % repr(override_uuid), file=error_file, flush=True)
        
        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        kn: KgtkNtriples = KgtkNtriples(
            input_file_paths=input_file_paths,
            output_file_path=output_kgtk_file,
            reject_file_path=reject_file_path,
            updated_namespace_file_path=updated_namespace_kgtk_file,
            namespace_file_path=namespace_kgtk_file,
            namespace_id_prefix=namespace_id_prefix,
            namespace_id_use_uuid=namespace_id_use_uuid,
            namespace_id_counter=namespace_id_counter,
            namespace_id_zfill=namespace_id_zfill,
            output_only_used_namespaces=output_only_used_namespaces,
            build_new_namespaces=build_new_namespaces,
            newnode_prefix=newnode_prefix,
            newnode_use_uuid=newnode_use_uuid,
            newnode_counter=newnode_counter,
            newnode_zfill=newnode_zfill,
            allow_lax_uri=allow_lax_uri,
            allow_unknown_datatype_iris=allow_unknown_datatype_iris,
            allow_turtle_quotes=allow_turtle_quotes,
            allow_lang_string_datatype=allow_lang_string_datatype,
            lang_string_tag=lang_string_tag,
            local_namespace_prefix=local_namespace_prefix,
            local_namespace_use_uuid=local_namespace_use_uuid,
            prefix_expansion_label=prefix_expansion_label,
            structured_value_label=structured_value_label,
            structured_uri_label=structured_uri_label,
            build_id=build_id,
            build_datatype_column=build_datatype_column,
            datatype_column_name=datatype_column_name,
            validate=validate,
            summary=summary,
            override_uuid=override_uuid,
            idbuilder_options=idbuilder_options,
            reader_options=reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose)

        kn.process()
    
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




import sys
import torch
import typing
import logging
from typing import List
from pathlib import Path
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.exceptions import KGTKException
from sentence_transformers import SentenceTransformer
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions


class EmbeddingVector:
    def __init__(self,
                 model_name=None,
                 output_property_name: str = "text_embedding",
                 sentence_property_name: str = "sentence",
                 output_format: str = "kgtk"):
        self._logger = logging.getLogger(__name__)
        if not model_name:
            self.model_name = 'bert-base-nli-mean-tokens'
        else:
            self.model_name = model_name
        self._logger.info("Using model {}".format(self.model_name))
        if torch.cuda.is_available():
            self.model = SentenceTransformer(self.model_name, device=('cuda:0'))
        else:
            self.model = SentenceTransformer(self.model_name)

        self.selected_gpu_device_index = 0
        self.total_gpu_count = torch.cuda.device_count()
        self.column_names = ['node1', 'label', 'node2']
        self.error_file = sys.stderr
        self.output_property_name = output_property_name
        self.sentence_property_name = sentence_property_name
        self.output_format = output_format

    def retry_next_gpu(self, error):
        self.selected_gpu_device_index += 1
        if self.selected_gpu_device_index >= self.total_gpu_count:
            self._logger.error("Attempted task on all available GPUs")
            raise error
        else:
            self.model = SentenceTransformer(self.model_name, device=('cuda:' + str(self.selected_gpu_device_index)))
            self._logger.debug(f"Reattempting task on GPU device: {('cuda:' + str(self.selected_gpu_device_index))}")

    def get_sentences_embedding(self, sentences: typing.List[str]):
        """
            transform a list of sentences to embedding vectors
        """
        while True:
            # Re-attempt executing the model on a different GPU until all GPU's have been tried
            try:
                sentence_embeddings = self.model.encode(sentences, show_progress_bar=True)
                break  # If there is no error, this will break the loop and return the results
            except RuntimeError as e:
                self.retry_next_gpu(e)
        return sentence_embeddings

    def process_sentences_kgtk(self,
                               input_file_path: Path,
                               output_file_path: Path,
                               error_file: typing.TextIO = sys.stderr,
                               reader_options: typing.Optional[KgtkReaderOptions] = None,
                               value_options: typing.Optional[KgtkValueOptions] = None,
                               verbose: bool = False,
                               batch_size: int = 100000
                               ):
        kr: KgtkReader = KgtkReader.open(input_file_path,
                                         error_file=error_file,
                                         options=reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         )
        if kr.node1_column_idx < 0:
            raise KGTKException("Missing column: node1 or alias")
        if kr.label_column_idx < 0:
            raise KGTKException("Missing column: label or alias")
        if kr.node2_column_idx < 0:
            raise KGTKException("Missing column: node2 or alias")

        self._logger.debug("node1 column index = {}".format(kr.node1_column_idx))
        self._logger.debug("label column index = {}".format(kr.label_column_idx))
        self._logger.debug("node2 column index = {}".format(kr.node2_column_idx))

        sentences = []
        qnodes = []

        all_qnodes = []
        all_vectors = []

        if self.output_format == 'kgtk':
            kw: KgtkWriter = KgtkWriter.open(self.column_names,
                                             output_file_path,
                                             require_all_columns=False,
                                             error_file=self.error_file
                                             )
        else:
            kw = open(output_file_path, 'w')

        for row in kr:
            if 0 < batch_size == len(sentences):
                vectors = self.get_sentences_embedding(sentences)
                if self.output_format == 'kgtk':
                    self.write_vectors(kw, self.output_property_name, qnodes, vectors)
                else:
                    all_qnodes.extend(qnodes)
                    all_vectors.extend(vectors)
                sentences = []
                qnodes = []
            else:
                if row[kr.label_column_idx].strip() == self.sentence_property_name:
                    qnodes.append(row[kr.node1_column_idx])
                    sentences.append(row[kr.node2_column_idx])

        if len(qnodes) > 0 and len(sentences) > 0:
            vectors = self.get_sentences_embedding(sentences)
            if self.output_format == 'kgtk':
                self.write_vectors(kw, self.output_property_name, qnodes, vectors)
            else:
                all_qnodes.extend(qnodes)
                all_vectors.extend(vectors)

        if len(all_qnodes) > 0 and len(all_vectors) > 0 and self.output_format == 'w2v':
            kw.write(f'{len(all_qnodes)} {len(all_vectors[0])}\n')
            for qnode, vector in zip(all_qnodes, all_vectors):
                kw.write(f'{qnode} {" ".join(map(str, vector))}\n')

        kr.close()
        kw.close()

    @staticmethod
    def write_vectors(kw: KgtkWriter, property_name: str, qnodes: typing.List[str], vectors: List[float]):
        for qnode, vector in zip(qnodes, vectors):
            kw.write([qnode, property_name, ",".join(map(str, vector))])



"""
Normalize KGTK edge file by removing additional columns that match the "lift" pattern..
"""
from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles
from kgtk.lift.kgtklift import KgtkLift

LOWER_COMMAND: str = 'lower'
NORMALIZE_EDGES_COMMAND: str = 'normalize-edges'

def parser():
    return {
        'aliases': [ LOWER_COMMAND, NORMALIZE_EDGES_COMMAND ],
        'help': 'Normalize a KGTK edge file by reversing the "lift" pattern or converting secondary edge columns to new edges.',
        'description': 'Normalize a KGTK edge file by removing columns that match a "lift" pattern and converting remaining additional columns to new edges.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalue import KgtkValue
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert
    _command: str = parsed_shared_args._command
    _mode: str = parsed_shared_args._mode

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()
    parser.add_output_file(who="An optional output file for new edges (normalized and/or lowered). " +
                           "If omitted, new edges will go in the main output file.",
                           dest="new_edges_file",
                           options=["--new-edges-file"],
                           metavar="NEW_EDGES_FILE",
                           optional=True)


    parser.add_argument(      "--columns", "--columns-to-lower", "--columns-to-remove", action="store", type=str, dest="columns_to_lower", nargs='+',
                              help="Columns to lower and remove as a space-separated list. (default=all columns other than key columns)")

    parser.add_argument(      "--base-columns", dest="base_columns",
                              help=h("Optionally, explicitly list the base column for each column being lowered. " +
                              " --base-columns and --columns-to-lower must have the same number of entries."), nargs='*')

    parser.add_argument(      "--label-values", action="store", type=str, dest="label_values", nargs='*',
                              help=h("When not empty, a list of label values to use for lowered edges when --base-columns is used, overriding the original column names. (default=%(default)s)"))

    parser.add_argument(      "--lift-separator", dest="lift_separator",
                              help=h("The separator between the base column and the label value. (default=%(default)s)."),
                              default=KgtkLift.DEFAULT_OUTPUT_LIFTED_COLUMN_SEPARATOR)

    parser.add_argument(      "--ignore-empty-node1", dest="ignore_empty_node1",
                              help=h("When True, ignore attempts to lower into a new record with an empty node1 value. (default=%(default)s)"),
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--ignore-empty-node2", dest="ignore_empty_node2",
                              help=h("When True, ignore attempts to lower into a new record with an empty node2 value. (default=%(default)s)"),
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument(      "--add-id", dest="add_id",
                              help="When True, add an id column to the output (if not already present). (default=%(default)s)",
                              type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    if _command == LOWER_COMMAND:
        parser.add_argument(      "--lower", dest="lower",
                                  help=h("When True, lower columns that match a lift pattern. (default=%(default)s)"),
                                  type=optional_bool, nargs='?', const=True, default=True, metavar="True|False")
        
        parser.add_argument(      "--normalize", dest="normalize",
                                  help=h("When True, normalize columns that do not match a lift pattern. (default=%(default)s)"),
                                  type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    elif _command == NORMALIZE_EDGES_COMMAND:
        parser.add_argument(      "--lower", dest="lower",
                                  help=h("When True, lower columns that match a lift pattern. (default=%(default)s)"),
                                  type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")
        
        parser.add_argument(      "--normalize", dest="normalize",
                                  help=h("When True, normalize columns that do not match a lift pattern. (default=%(default)s)"),
                                  type=optional_bool, nargs='?', const=True, default=True, metavar="True|False")

    else:
        parser.add_argument(      "--lower", dest="lower",
                                  help="When True, lower columns that match a lift pattern. (default=%(default)s)",
                                  type=optional_bool, nargs='?', const=True, default=_command != NORMALIZE_EDGES_COMMAND, metavar="True|False")

        parser.add_argument(      "--normalize", dest="normalize",
                                  help="When True, normalize columns that do not match a lift pattern. (default=%(default)s)",
                                  type=optional_bool, nargs='?', const=True, default=_command != LOWER_COMMAND, metavar="True|False")

    parser.add_argument(      "--deduplicate-new-edges", dest="deduplicate_new_edges",
                              help="When True, deduplicate new edges. Not suitable for large files. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=True, metavar="True|False")

    KgtkIdBuilderOptions.add_arguments(parser, default_style=KgtkIdBuilderOptions.CONCAT_NLN_NUM_STYLE, expert=_expert)
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE if _mode == "NONE" else KgtkReaderMode.EDGE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        new_edges_file: KGTKFiles,

        base_columns: typing.Optional[typing.List[str]] = None,
        columns_to_lower: typing.Optional[typing.List[str]] = None,
        label_values: typing.Optional[typing.List[str]] = None,
        lift_separator: str = KgtkLift.DEFAULT_OUTPUT_LIFTED_COLUMN_SEPARATOR,
        ignore_empty_node1: bool = False,
        ignore_empty_node2: bool = False,
        add_id: bool = False,
        lower: bool = False,
        normalize: bool = False,
        deduplicate_new_edges: bool = True,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.reshape.kgtkidbuilder import KgtkIdBuilder, KgtkIdBuilderOptions
    from kgtk.value.kgtkvalue import KgtkValue
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    new_edges_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(new_edges_file, who="Label file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    idbuilder_options: KgtkIdBuilderOptions = KgtkIdBuilderOptions.from_dict(kwargs)
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file)
        if new_edges_kgtk_file is not None:
            print("--label-file=%s" % str(new_edges_kgtk_file), file=error_file)

        if base_columns is not None:
            print("--base-columns %s" % " ".join(base_columns), file=error_file)
        if columns_to_lower is not None:
            print("--columns-to-lower %s" % " ".join(columns_to_lower), file=error_file)
        if label_values is not None:
            print("--label-values %s" % " ".join(label_values), file=error_file)
        print("--lift-separator=%s" % lift_separator, file=error_file)
        print("--add-id=%s" % add_id, file=error_file)
        print("--lower=%s" % lower, file=error_file)
        print("--ignore-empty-node1=%s" % ignore_empty_node1, file=error_file)
        print("--ignore-empty-node2=%s" % ignore_empty_node2, file=error_file)
        print("--normalize=%s" % normalize, file=error_file)
        print("--deduplicate-labels=%s" % deduplicate_new_edges, file=error_file)

        idbuilder_options.show(out=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)


    if not lower and not normalize:
        raise KGTKException("One or both of --lower and --normalize must be requested.")

    try:
        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        # Map the index of a column being removed to the index of the base column that supplies its node1 value.
        lower_map: typing.MutableMapping[int, typing.Tuple[int, str]] = dict()

        node1_column_name: str = kr.get_node1_column_actual_name()
        label_column_name: str = kr.get_label_column_actual_name()
        node2_column_name: str = kr.get_node2_column_actual_name()
        id_column_name: str = kr.get_id_column_actual_name()

        key_column_names: typing.List[str] = list()
        key_column_idxs: typing.Set[int] = set()

        if node1_column_name != "":
            if verbose:
                print("Node1 column name: %s" % node1_column_name, file=error_file, flush=True)
            key_column_names.append(node1_column_name)
            key_column_idxs.add(kr.node1_column_idx)
            
        if label_column_name != "":
            if verbose:
                print("Label column name: %s" % label_column_name, file=error_file, flush=True)
            key_column_names.append(label_column_name)
            key_column_idxs.add(kr.label_column_idx)
            
        if node2_column_name != "":
            if verbose:
                print("Node2 column name: %s" % node2_column_name, file=error_file, flush=True)
            key_column_names.append(node2_column_name)
            key_column_idxs.add(kr.node2_column_idx)
            
        if id_column_name != "":
            if verbose:
                print("Id column name: %s" % id_column_name, file=error_file, flush=True)
            key_column_names.append(id_column_name)
            key_column_idxs.add(kr.id_column_idx)
        elif normalize:
            raise KGTKException("--normalize was requested but the ID column was not found.")
  
        base_name: str
        new_label_value: str
        column_name: str
        idx: int
        # There are three option patterns.

        if columns_to_lower is not None and len(columns_to_lower) > 0 and base_columns is not None and len(base_columns) > 0:
            # Pattern 1: len(columns_to_lower) > 0 and len(base_columns) == len(columns_to_lower)
            # column_names and base_columns are paired. New records use label_values if specified.
            if len(columns_to_lower) != len(base_columns):
                raise KGTKException("There are %d columns to lower but only %d base columns." % (len(columns_to_lower), len(base_columns)))

            if label_values is not None and len(label_values) > 0 and len(label_values) != len (columns_to_lower):
                raise KGTKException("There are %d columns to lower but only %d label values." % (len(columns_to_lower), len(label_values)))

            for idx, column_name in enumerate(columns_to_lower):
                base_name = base_columns[idx]
                if column_name not in kr.column_names:
                    raise KGTKException("Column %s is an unknown column, cannot remove it." % repr(column_name))

                if column_name in key_column_names:
                    raise KGTKException("Column %s is a key column, cannot remove it." % repr(column_name))

                if base_name not in kr.column_names:
                    raise KGTKException("For column name %s, base name %s is unknown" % (repr(column_name), repr(base_name)))

                if normalize and base_name == id_column_name:
                    lower_map[kr.column_name_map[column_name]] = (kr.column_name_map[base_name], column_name)
                else:
                    if not lower:
                        raise KGTKException("--lower is not enabled for column %s, base name %s" % (repr(column_name), repr(base_name)))
                    if label_values is not None and len(label_values) > 0 and len(label_values[idx]) > 0:
                        lower_map[kr.column_name_map[column_name]] = (kr.column_name_map[base_name], label_values[idx])
                    else:
                        lower_map[kr.column_name_map[column_name]] = (kr.column_name_map[base_name], column_name)

        elif columns_to_lower is not None and len(columns_to_lower) > 0 and (base_columns is None or len(base_columns) == 0):
            # Pattern 2: len(columns_to_lower) > 0 and len(base_columns) == 0
            # Each column name is split at the lift separator to determine the base name and label value.
            if len(lift_separator) == 0:
                raise KGTKException("The --lift-separator must not be empty.")

            for idx, column_name in enumerate(columns_to_lower):
                if column_name not in kr.column_names:
                    raise KGTKException("Column %s is an unknown column, cannot remove it." % repr(column_name))

                if column_name in key_column_names:
                    raise KGTKException("Column %s is a key column, cannot remove it." % repr(column_name))

                if lower and lift_separator in column_name:
                    base_name, new_label_value = column_name.split(lift_separator, 1)
                    if base_name not in kr.column_names:
                        raise KGTKException("For column name %s, base name %s is not known" % (repr(column_name), repr(base_name)))

                elif normalize:
                    base_name = id_column_name
                    new_label_value = column_name

                else:
                    raise KGTKException("Unable to parse column name %s, no separator (%s)." % (repr(column_name), repr(lift_separator)))

                lower_map[kr.column_name_map[column_name]] = (kr.column_name_map[base_name], new_label_value)

        elif columns_to_lower is None or len(columns_to_lower) == 0:
            # Pattern 3: len(columns_to_lower) == 0.
            # Any column that matches a lift pattern against one of the
            # key columns (node1, label, node2, id, or their aliases)
            # will be lowered.
            if len(lift_separator) == 0:
                raise KGTKException("The --lift-separator must not be empty.")

            if base_columns is None or len(base_columns) == 0:
                # The base name list wasn't supplied.  Use [node1, label, node2, id]
                base_columns = list(key_column_names)
                if verbose:
                    print("Using the default base columns: %s" % " ".join(base_columns), file=error_file, flush=True)
            else:
                if verbose:
                    print("Using these base columns: %s" % " ".join(base_columns), file=error_file, flush=True)

            for idx, column_name in enumerate(kr.column_names):
                # Skip the node1, label, node12, and id columns
                if idx in key_column_idxs:
                    if verbose:
                        print("column %s is a key column, skipping." % repr(column_name), file=error_file, flush=True)
                    continue

                # Does this column match a lifting pattern?
                if lower and lift_separator in column_name:
                    base_name, new_label_value = column_name.split(lift_separator, 1)

                    if base_name not in base_columns:
                        if verbose:
                            print("Column %s contains base name %s, which is not a base column." % (repr(column_name), repr(base_name)),  file=error_file, flush=True)
                        continue
                                  
                elif normalize:
                    base_name = id_column_name
                    new_label_value = column_name

                else:
                    if verbose:
                        print("Column %s does not contain the separator %s and not normalizing, skipping." % (repr(column_name), repr(lift_separator)), file=error_file, flush=True)
                    continue

                # This test should be redundant.
                if base_name in kr.column_names:
                    lower_map[idx] = (kr.column_name_map[base_name], new_label_value)
                else:
                    raise KGTKException("Base name %s was unexpectedly not found." % repr(base_name))

        if len(lower_map) == 0:
            raise KGTKException("There are no columns to lower or normalize.")

        if verbose:
            print("The following columns will be lowered or normalized", file=error_file, flush=True)
            for idx in sorted(lower_map.keys()):
                column_name = kr.column_names[idx]
                base_idx, new_label_value = lower_map[idx]
                base_name = kr.column_names[base_idx]
                print(" %s from %s (label %s)" % (column_name, base_name, repr(new_label_value)), file=error_file, flush=True)

        output_column_names: typing.List[str] = list()
        for idx, column_name in enumerate(kr.column_names):
            if idx not in lower_map:
                output_column_names.append(column_name)

        # Create the ID builder.
        idb: typing.Optional[KgtkIdBuilder] = None
        if add_id:
            idb = KgtkIdBuilder.from_column_names(output_column_names, idbuilder_options)
            output_column_names = idb.column_names.copy()

        if verbose:
            print("The output columns are: %s" % " ".join(output_column_names), file=error_file, flush=True)

        if verbose:
            print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(output_column_names,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns = False, # Simplifies writing the labels
                                         verbose=verbose,
                                         very_verbose=very_verbose)
        shuffle_list: typing.List[int] = kw.build_shuffle_list(kr.column_names)

        lkw: typing.Optional[KgtkWriter] = None
        if new_edges_kgtk_file is not None:
            if verbose:
                print("Opening the label output file: %s" % str(new_edges_kgtk_file), file=error_file, flush=True)

            label_column_names = [ node1_column_name, label_column_name, node2_column_name ]                
            lkw = KgtkWriter.open(label_column_names,
                                  new_edges_kgtk_file,
                                  mode=KgtkWriter.Mode.EDGE,
                                  verbose=verbose,
                                  very_verbose=very_verbose)
                      

        # Optionally deduplicate the labels
        #  set(node1_value + KgtkFormat.SEPARATOR + node2_value)
        label_set: typing.Set[str] = set()
        label_key: str

        input_line_count: int = 0
        output_line_count: int = 0
        label_line_count: int = 0
        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            output_row: typing.List[str] = kw.shuffle(row, shuffle_list=shuffle_list)
            kw.write(output_row)
            output_line_count += 1

            id_seq_num: int = 0
            column_idx: int
            for column_idx in lower_map.keys():
                node1_idx: int
                node1_idx, new_label_value = lower_map[column_idx]
                node1_value: str
                node1_value = row[node1_idx]
                if len(node1_value) == 0:
                    if ignore_empty_node1:
                        continue # TODO: raise an exception
                    else:
                        raise KGTKException("Empty node1 value when lowering %d to %d: %s in input line %d" % (column_idx,
                                                                                                               node1_idx,
                                                                                                               new_label_value,
                                                                                                               input_line_count))

                item: str = row[column_idx]
                if len(item) == 0:
                    if ignore_empty_node2:
                        continue # Ignore empty node2 values.
                    else:
                        raise KGTKException("Empty node2 value when lowering %d to %d: %s in input line %d" % (column_idx,
                                                                                                               node1_idx,
                                                                                                               new_label_value,
                                                                                                               input_line_count))

                # Ths item might be a KGTK list.  Let's split it, because
                # lists aren't allow in the node2 values we'll generate.
                node2_value: str
                for node2_value in KgtkValue.split_list(item):
                    if len(node2_value) == 0:
                        if ignore_empty_node2:
                            continue # Ignore empty node2 values in a list.
                        else:
                            raise KGTKException("Empty node2 value in a list when lowering %d to %d: %s in input line %d" % (column_idx,
                                                                                                                             node1_idx,
                                                                                                                             new_label_value,
                                                                                                                             input_line_count))

                    if deduplicate_new_edges:
                        label_key = node1_value + KgtkFormat.KEY_FIELD_SEPARATOR + new_label_value + KgtkFormat.KEY_FIELD_SEPARATOR + node2_value
                        if label_key in label_set:
                            continue
                        else:
                            label_set.add(label_key)

                    lowered_input_row: typing.List[str] = ["" for idx in range(kr.column_count)]
                    lowered_input_row[kr.node1_column_idx] = node1_value
                    lowered_input_row[kr.label_column_idx] = new_label_value
                    lowered_input_row[kr.node2_column_idx] = node2_value

                    lowered_output_row: typing.List[str] = kw.shuffle(lowered_input_row, shuffle_list=shuffle_list)
                    if idb is not None:
                        id_seq_num += 0
                        lowered_output_row = idb.build(lowered_output_row, id_seq_num, already_added=True)
                    if lkw is not None:
                        lkw.write(lowered_output_row)
                        label_line_count += 1
                    else:
                        kw.write(lowered_output_row)
                        label_line_count += 1
                        output_line_count += 1

        if verbose:
            print("Read %d rows, wrote %d rows with %d labels." % (input_line_count, output_line_count, label_line_count), file=error_file, flush=True)

        kw.close()
        if lkw is not None:
            lkw.close()

        return 0

    except Exception as e:
        kgtk_exception_auto_handler(e)
        return 1



import datetime
import importlib
from io import StringIO
import itertools
import os
import pkgutil
import sh # type: ignore
import shutil # type: ignore
import signal
import sys
import time
import typing

from kgtk import cli
from kgtk.exceptions import KGTKException, KGTKExceptionHandler, KGTKArgumentParseException
from kgtk import __version__
from kgtk.cli_argparse import KGTKArgumentParser, add_shared_arguments, add_default_arguments, CheckDepsAction


# module name should NOT start with '__' (double underscore)
handlers = [(x.name, "kgtk.cli") for x in pkgutil.iter_modules(cli.__path__)
                   if not x.name.startswith('__')]

try:
    from kgtk_extensions import cli as cliext
    ext_handlers = [(x.name, "kgtk_extensions.cli") for x in pkgutil.iter_modules(cliext.__path__)
                    if not x.name.startswith('__')]
    handlers = sorted(handlers + ext_handlers)
except ImportError:
    pass

# import signal
# signal.signal(signal.SIGPIPE, signal.SIG_DFL)

pipe_delimiter = '/'
sequential_delimiter = '/.'
parallel_delimiter = '//'
ret_code = 0

def cmd_done(cmd, success, exit_code):
    # cmd.cmd -> complete command line
    global ret_code
    ret_code = exit_code


_save_progress: bool = False
_save_progress_debug: bool = False
_save_progress_tty: typing.Optional[str] = None
_save_progress_command: typing.Optional[typing.Any] = None
def progress_startup(pid: typing.Optional[int] = None, fd: typing.Optional[int] = None):
    # This can be called multiple times, if it desired to monitor several
    # input files in sequence.
    #
    # If pid is None, the current process will be monitored.  If pid is not
    # None, the specified process will be monitored.
    #
    # If target_fd is None, them all fds will be monitored.  If target is not
    # None, the specific fd will be monitored: it must already be open, and it
    # should be an input file.  There is no option to moitor multiple specific
    # input files other than calling this routine sequentially
    #
    # TODO: use the envar KGTK_PV_COMMAND to get the `pv` command.
    global _save_progress, _save_progress_tty, _save_progress_debug
    if _save_progress and _save_progress_tty is not None:
        global _save_progress_command
        if _save_progress_command is not None:
            # Shut down an existing process monitor.
            try:
                _save_progress_command.terminate()
            except Exception:
                pass
            _save_progress_command = None

        # Give up if cannot find `pv`:
        if shutil.which('pv') is None:
            if _save_progress_debug:
                print("progress_startup: cannot find pv.", file=sys.stderr, flush=True)
            return

        # Start a process monitor.
        if pid is None:
            pid = os.getpid()
        try:
            if fd is None:
                if _save_progress_debug:
                    print("progress_startup: starting pv with pid %d" % pid, file=sys.stderr, flush=True)
                _save_progress_command = sh.pv("-d {}".format(pid), _out=_save_progress_tty, _err=_save_progress_tty, _bg=True)
            else:
                if _save_progress_debug:
                    print("progress_startup: starting pv with pid %d fd %d" % (pid, fd), file=sys.stderr, flush=True)
                _save_progress_command = sh.pv("-d {}:{}".format(pid, fd),
                                               _out=_save_progress_tty, _err=_save_progress_tty, _bg=True)
        except Exception as e:
            # Ignore the exception unless _save_progress_debug is True.
            if _save_progress_debug:
                print("progress_startup: %s" % str(e), file=sys.stderr, flush=True)
                raise

def progress_shutdown():
    global _save_progress_command
    if _save_progress_command is not None:
        try:
            _save_progress_command.terminate()
        except Exception:
            pass
        _save_progress_command = None
    
def split_list(sequence, sep):
    """From stack overflow;
    https://stackoverflow.com/questions/54372218/how-to-split-a-list-into-sublists-based-on-a-separator-similar-to-str-split
    """
    chunk = []
    for val in sequence:
        if val == sep:
            yield chunk
            chunk = []
        else:
            chunk.append(val)
    yield chunk

def cli_single_command(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    import copy
    if parsed_shared_args._pipedebug:
        print("pid %d: Executing a single command: %s" % (os.getpid(), repr(args)), file=sys.stderr, flush=True)
    ret_code: int = 0
    cmd_args = copy.deepcopy(args)
    cmd_name = cmd_args[0].replace("_", "-")
    cmd_args[0] = cmd_name
    # build sub-parser
    if cmd_name in subparser_lookup:
        mod, sub_parser = subparser_lookup[cmd_name]
        add_default_arguments(sub_parser)  # call this before adding other arguments
        if cmd_name not in subparsers_built:
            if hasattr(mod, 'add_arguments_extended'):
                psa = copy.deepcopy(parsed_shared_args)
                psa._command = cmd_name
                mod.add_arguments_extended(sub_parser, psa)
            else:
                mod.add_arguments(sub_parser)
            subparsers_built.add(cmd_name)
    parsed_args = parser.parse_args(cmd_args)

    # load module
    kwargs = {}
    func = None
    if parsed_args.cmd:
        h = parsed_args.cmd
        func = mod.run

        # remove sub-command name
        kwargs = vars(parsed_args)
        del kwargs['cmd']

        # set shared arguments
        for sa in vars(parsed_shared_args):
            # Shared arguments that have been accepted by the subcommand
            # (see KGTKArgumentParser.accept_shared_argument(...))
            # will be passed as keyword arguments to the subcommand.
            if sa not in sub_parsers.choices[h].shared_arguments:
                del kwargs[sa]
            else:
                kwargs[sa] = getattr(parsed_shared_args, sa)

    global _save_progress
    _save_progress = parsed_shared_args._progress
    global _save_progress_debug
    _save_progress_debug = parsed_shared_args._progress_debug
    global _save_progress_tty
    _save_progress_tty = parsed_shared_args._progress_tty
    if parsed_shared_args._progress:
        if hasattr(mod, 'custom_progress') and mod.custom_progress():
            pass
        else:
            progress_startup()

    # run module
    try: 
      kgtk_exception_handler = KGTKExceptionHandler(debug=parsed_shared_args._debug)
      ret_code = kgtk_exception_handler(func, **kwargs)
    except KeyboardInterrupt as e:
        print("\nKeyboard interrupt in %s." % " ".join(args), file=sys.stderr, flush=True)
        progress_shutdown()
        if hasattr(mod, 'keyboard_interrupt'):
            mod.keyboard_interrupt()

        # Silently exit instead of re-raising the KeyboardInterrupt.
        # raise
    return ret_code

def cli_piped_commands(parallel_pipes, args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    ret_code: int = 0
    if parsed_shared_args._pipedebug:
        print("pid %d: Building a KGTK pipe" % (os.getpid()), file=sys.stderr, flush=True)
    processes = [ ]
    try:
        # TODO: Restore the prior signal handler when done.
        def sigterm_handler(signal, frame):
            # This handles pipe shutdowns.
            if parsed_shared_args._pipedebug:
                print("\npipe: sigterm_handler", file=sys.stderr, flush=True)
            raise KeyboardInterrupt

        signal.signal(signal.SIGTERM, sigterm_handler)

        for parallel_idx, pipe in enumerate(parallel_pipes):
            if parallel_idx > 0 and parsed_shared_args._pipedebug:
                print("*** in parallel with ***", file=sys.stderr, flush=True)

            prior_process = None
            for pipe_idx, cmd_args in enumerate(pipe):
                # add shared arguments
                full_args = [ ]
                for shared_arg in shared_args:
                    if (parallel_idx == 0 and pipe_idx == 0) or str(shared_arg) != "--progress":
                        full_args.append(shared_arg)
                full_args.extend(cmd_args)

                kwargs = {
                    "_bg_exc": False,
                    "_done": cmd_done,
                    "_err": sys.stderr,
                    "_bg": True,
                    "_internal_bufsize": 1,
                    "_new_session": False,
                }

                # add specific arguments
                if pipe_idx == 0:  # The first command reads from our STDIN.
                    kwargs["_in"] = sys.stdin

                if pipe_idx + 1 < len(pipe):
                    # All commands but the last pipe their output to the next command.
                    kwargs["_piped"] = True
                else:
                    # The last command writes to our STDOUT.
                    kwargs["_out"] = sys.stdout

                if parsed_shared_args._pipedebug:
                    cmd_str = " ".join(full_args)
                    for key in kwargs:
                        cmd_str += " " + key + "=" + str(kwargs[key])
                    print("pipe[%d][%d]: kgtk %s" % (parallel_idx, pipe_idx, cmd_str), file=sys.stderr, flush=True)

                if prior_process is None:
                    new_process = sh.kgtk(*full_args, **kwargs)
                else:
                    new_process = sh.kgtk(prior_process, *full_args, **kwargs)
                processes.append(new_process)
                prior_process = new_process                
                
        if parsed_shared_args._pipedebug:
            print("*** waiting ***", file=sys.stderr, flush=True)
        for process in processes:
            process.wait()

    except sh.SignalException_SIGPIPE:
        if parsed_shared_args._pipedebug:
            print("\npipe: sh.SignalException_SIGPIPE", file=sys.stderr, flush=True)

    except sh.SignalException_SIGTERM:
        if parsed_shared_args._pipedebug:
            print("\npipe: sh.SignalException_SIGTERM", file=sys.stderr, flush=True)
        if len(processes) > 0:
            for process in processes:
                try:
                    process.terminate()
                except Exception:
                    pass
        # raise

    except KeyboardInterrupt as e:
        if parsed_shared_args._pipedebug:
            print("\npipe: KeyboardInterrupt", file=sys.stderr, flush=True)
        if len(processes) > 0:
            for idx in range(len(processes)):
                process = processes[idx]
                pgid = process.pgid
                print("Killing cmd %d process group %d" % (idx, pgid), file=sys.stderr, flush=True)
                try:
                    process.signal_group(signal.SIGINT)
                except Exception:
                    pass
        return -1

    except sh.ErrorReturnCode as e:
        if parsed_shared_args._pipedebug:
            print("\npipe: sh.ErrorReturnCode", file=sys.stderr, flush=True)
        # mimic parser exit
        parser.exit(KGTKArgumentParseException.return_code, e.stderr.decode('utf-8'))
    return ret_code

def cli_entry_pipe(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    if pipe_delimiter not in args and parallel_delimiter not in args:
        return cli_single_command(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)

    # parse internal pipe
    parallel_pipes: typing.List[typing.List[typing.List[str]]] = list()
    pipe: typing.List[typing.List[str]] = list()
    cmd_args: typing.List[str] = list()
    for arg in args:
        if arg == parallel_delimiter:
            if len(cmd_args) > 0:
                pipe.append(cmd_args)
                cmd_args = list()
            if len(pipe) > 0:
                parallel_pipes.append(pipe)
                pipe = list()
        elif arg == pipe_delimiter:
            if len(cmd_args) > 0:
                pipe.append(cmd_args)
                cmd_args = list()
        else:
            cmd_args.append(arg)
    if len(cmd_args) > 0:
        pipe.append(cmd_args)
    if len(pipe) > 0:
        parallel_pipes.append(pipe)

    if len(parallel_pipes) == 0:
        parser.print_usage()
        parser.exit(KGTKArgumentParseException.return_code)
    return cli_piped_commands(parallel_pipes, args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)
    
def cli_entry_sequential_commands(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)->int:
    # parse internal sequence of pipes
    ret_code: int = 0
    for sequential_idx, commands in enumerate(split_list(args, sequential_delimiter)):
        if sequential_idx > 0 and parsed_shared_args._pipedebug:
            print("*** followed by ***", file=sys.stderr, flush=True)
        ret_code = cli_entry_pipe(commands, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)
        if ret_code != 0:
            break

    return ret_code

def cli_entry(*args):
    """
    Usage:
        kgtk <command> [options]
    """
    global ret_code

    # Capture the initial time for timing measurements.
    start_time: float = time.time()
    process_start_time: float = time.process_time()

    # get all arguments
    if not args:
        args = tuple(sys.argv)
    args = args[1:]

    # base parser for shared arguments
    base_parser = KGTKArgumentParser(add_help=False)
    base_parser.add_argument(
        '-V', '--version',
        action='version',
        version='KGTK %s' % __version__,
        help='show KGTK version number and exit.'
    )
    base_parser.add_argument(
        '--check-deps',
        action=CheckDepsAction,
        help='check dependencies',
    )
    shared_args = base_parser.add_argument_group('shared optional arguments')
    shared_args.add_argument('--debug', dest='_debug', action='store_true',
                             default=os.getenv('KGTK_OPTION_DEBUG', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable debug mode')
    
    shared_args.add_argument('--kgtkmode', dest='_mode', action='store',
                             default=os.getenv('KGTK_OPTION_KGTK_MODE', 'AUTO').upper(),
                             choices=['NONE', 'EDGE', 'NODE', 'AUTO' ],
                             help='KGTK file reading mode (default=AUTO)')
    
    shared_args.add_argument('--expert', dest='_expert', action='store_true',
                             default=os.getenv('KGTK_OPTION_EXPERT', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable expert mode')
    
    shared_args.add_argument('--pipedebug', dest='_pipedebug', action='store_true',
                             default=os.getenv('KGTK_OPTION_PIPEDEBUG', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable pipe debug mode')
    
    shared_args.add_argument('--progress', dest='_progress', action='store_true',
                             default=os.getenv('KGTK_OPTION_PROGRESS', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable progress monitoring')
    
    shared_args.add_argument('--progress-debug', dest='_progress_debug', action='store_true',
                             default=os.getenv('KGTK_OPTION_PROGRESSDEBUG', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable progress debug mode, which will show exceptions occuring during progress monitoring startup')
    
    shared_args.add_argument('--progress-tty', dest='_progress_tty', action='store',
                             default=os.getenv('KGTK_OPTION_PROGRESS_TTY', "/dev/tty"),
                             help='progress monitoring output tty')
    
    shared_args.add_argument('--timing', dest='_timing', action='store_true',
                             default=os.getenv('KGTK_OPTION_TIMING', 'False').lower() in ['y', 'yes', 'true'],
                             help='enable timing measurements')
    
    add_shared_arguments(shared_args)

    # parse shared arguments
    parsed_shared_args, rest_args = base_parser.parse_known_args(args)
    shared_args = tuple(filter(lambda a: a not in rest_args, args))
    args = tuple(rest_args)

    # complete parser, load sub-parser of each module
    parser = KGTKArgumentParser(
        parents=[base_parser], prog='kgtk',
        description='kgtk --- Knowledge Graph Toolkit',
    )
    sub_parsers = parser.add_subparsers(
        metavar='command',
        dest='cmd'
    )
    subparsers_built = set()
    subparser_lookup = {}
    sub_parsers.required = True
    for h in handlers:
        hname, hpath = h
        mod = importlib.import_module('.{}'.format(hname), hpath)
        subp = mod.parser()
        # only create sub-parser with sub-command name and defer full build
        cmd: str = hname.replace("_", "-")
        sub_parser = sub_parsers.add_parser(cmd, **subp)
        subparser_lookup[cmd] = (mod, sub_parser)
        if 'aliases' in subp:
            for alias in subp['aliases']:
                subparser_lookup[alias] = (mod, sub_parser)

    # add root level usage after sub-parsers are created
    # this won't pollute help info in sub-parsers
    parser.usage = '%(prog)s [options] command [ / command]*'

    ret_code = cli_entry_sequential_commands(args, parsed_shared_args, shared_args, parser, sub_parsers, subparser_lookup, subparsers_built)

    if parsed_shared_args._timing:
        end_time: float = time.time()
        elapsed_seconds: float = end_time - start_time

        process_end_time: float = time.process_time()
        process_elapsed_seconds: float = process_end_time - process_start_time

        cpu_ratio: float = process_elapsed_seconds / elapsed_seconds

        print("Timing: elapsed=%s CPU=%s (%5.1f%%): %s" % (str(datetime.timedelta(seconds=elapsed_seconds)),
                                                           str(datetime.timedelta(seconds=process_elapsed_seconds)),
                                                           cpu_ratio * 100.0,
                                                           " ".join(args)), file=sys.stderr, flush=True)

    return ret_code




"""
Compute paths between nodes in a KGTK graph.

TODO: Add --output-file
"""
from argparse import Namespace
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles


def parser():
    return {
        'help': 'Compute paths between nodes in a KGTK graph.'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True)
    parser.add_output_file()
    parser.add_input_file(who="KGTK file with path start and end nodes.",
                          options=["--path-file", "--path_file"], dest="path_file", metavar="PATH_FILE", optional=False)

    parser.add_argument('--statistics-only', dest='statistics_only',
                        help='If this flag is set, output only the statistics edges. Else, append the statistics to the original graph. (default=%(default)s)',
                        type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument('--undirected', dest="undirected",
                        help="Is the graph undirected or not? (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=False, metavar="True|False")

    parser.add_argument('--max-hops', '--max_hops', action="store", type=int, dest="max_hops", help="Maximum number of hops allowed.")

    parser.add_argument("--path-source", action="store", type=str, dest="source_column_name",
                        help='Name of the source column in the path file. (default: node1 or its alias)')
    parser.add_argument("--path-target", action="store", type=str, dest="target_column_name",
                        help='Name of the source column in the path file. (default: node2 or its alias)')

    parser.add_argument("--shortest-path", dest="shortest_path", metavar="True|False",
                        help="When true, shortest paths are returned. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="path", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        path_file: KGTKFiles,
        output_file: KGTKFiles,
        statistics_only: bool,
        undirected: bool,
        max_hops: int,

        source_column_name: typing.Optional[str],
        target_column_name: typing.Optional[str],
        shortest_path: bool,

        errors_to_stdout: bool,
        errors_to_stderr: bool,
        show_options: bool,
        verbose: bool,
        very_verbose: bool,

        **kwargs,  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ):
    # import modules locally
    from pathlib import Path
    import sys

    from graph_tool.all import find_vertex
    from graph_tool.topology import all_paths
    from graph_tool.topology import all_shortest_paths

    from kgtk.gt.gt_load import load_graph_from_kgtk
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    from kgtk.exceptions import KGTKException
    try:

        # Select where to send error messages, defaulting to stderr.
        error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

        # Build the option structures.
        input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
        path_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="path", fallback=True)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

        input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
        path_kgtk_file: Path = KGTKArgumentParser.get_input_file(path_file)
        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

        id_col = 'name'

        if verbose:
            print("Reading the path file: %s" % str(path_kgtk_file), file=error_file, flush=True)
        pairs = []
        pkr: KgtkReader = KgtkReader.open(path_kgtk_file,
                                          error_file=error_file,
                                          options=path_reader_options,
                                          value_options=value_options,
                                          verbose=verbose,
                                          very_verbose=very_verbose,
                                          )
        path_source_idx: int = pkr.get_node1_column_index(source_column_name)
        if path_source_idx < 0:
            print("Missing node1 (source) column name in the path file.", file=error_file, flush=True)

        path_target_idx: int = pkr.get_node2_column_index(target_column_name)
        if path_target_idx < 0:
            print("Missing node1 (target) column name in the path file.", file=error_file, flush=True)
        if path_source_idx < 0 or path_target_idx < 0:
            pkr.close()
            raise KGTKException("Exiting due to missing columns.")

        paths_read: int = 0
        path_row: typing.List[str]
        for path_row in pkr:
            paths_read += 1
            if len(path_row) != pkr.column_count:
                raise KGTKException("Exiting because line %d in the path file (%s) is the wrong length: %d columns expected, %d were read." % (paths_read,
                                                                                                                                               str(path_kgtk_file),
                                                                                                                                               pkr.column_count,
                                                                                                                                               len(path_row)))
            src: str = path_row[path_source_idx]
            tgt: str = path_row[path_target_idx]
            pairs.append((src, tgt))
        pkr.close()
        if verbose:
            print("%d path rows read" % paths_read, file=error_file, flush=True)
        if len(pairs) == 0:
            print("No path pairs found, the output will be empty.", file=error_file, flush=True)
        elif verbose:
            print("%d path pairs found" % len(pairs), file=error_file, flush=True)

        if verbose:
            print("Reading the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         error_file=error_file,
                                         options=input_reader_options,
                                         value_options=value_options,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
                                         )

        sub_index: int = kr.get_node1_column_index()
        if sub_index < 0:
            print("Missing node1 (subject) column.", file=error_file, flush=True)
        pred_index: int = kr.get_label_column_index()
        if pred_index < 0:
            print("Missing label (predicate) column.", file=error_file, flush=True)
        obj_index: int = kr.get_node2_column_index()
        if obj_index < 0:
            print("Missing node2 (object) column", file=error_file, flush=True)
        id_index: int = kr.get_id_column_index()
        if id_index < 0:
            print("Missing id column", file=error_file, flush=True)
        if sub_index < 0 or pred_index < 0 or obj_index < 0 or id_index < 0:
            kr.close()
            raise KGTKException("Exiting due to missing columns.")

        predicate: str = kr.column_names[pred_index]
        id_col_name: str = kr.column_names[id_index]

        G = load_graph_from_kgtk(kr, directed=not undirected, ecols=(sub_index, obj_index), verbose=verbose,
                                 out=error_file)

        output_columns: typing.List[str] = ['node1', 'label', 'node2', 'id']
        kw: KgtkWriter = KgtkWriter.open(output_columns,
                                         output_kgtk_file,
                                         mode=KgtkWriter.Mode.EDGE,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         verbose=verbose,
                                         very_verbose=very_verbose)

        id_count = 0
        if not statistics_only:
            for e in G.edges():
                sid, oid = e
                lbl = G.ep[predicate][e]
                kw.write(
                    [G.vp[id_col][sid], lbl, G.vp[id_col][oid], '{}-{}-{}'.format(G.vp[id_col][sid], lbl, id_count)])
                id_count += 1
            if verbose:
                print("%d edges found." % id_count, file=error_file, flush=True)

        id_count = 0
        path_id = 0
        for pair in pairs:
            source_node, target_node = pair
            source_ids = find_vertex(G, prop=G.properties[('v', id_col)], match=source_node)
            target_ids = find_vertex(G, prop=G.properties[('v', id_col)], match=target_node)
            if len(source_ids) == 1 and len(target_ids) == 1:
                source_id = source_ids[0]
                target_id = target_ids[0]
                if shortest_path:
                    _all_paths = all_shortest_paths(G, source_id, target_id, edges=True)
                else:
                    _all_paths = all_paths(G, source_id, target_id, cutoff=max_hops, edges=True)

                for path in _all_paths:
                    for edge_num, an_edge in enumerate(path):
                        edge_id = G.properties[('e', 'id')][an_edge]
                        node1: str = 'p%d' % path_id
                        kw.write([node1, str(edge_num), edge_id, '{}-{}-{}'.format(node1, edge_num, id_count)])
                        id_count += 1
                    path_id += 1

        if verbose:
            print("%d paths contining %d edges found." % (path_id, id_count), file=error_file, flush=True)

        kw.close()
        kr.close()

    except Exception as e:
        raise KGTKException('Error: ' + str(e))



"""This file defines the `kgtk(...)` function, which simplifies KGTK command
usage in JupyterLab.  The `kypher(...)` command is also defined for convenience.

Importing
=========


from kgtk.functions import kgtk, kypher

"""
import csv
from io import StringIO
from IPython.core.display import display, HTML, JSON, Markdown
import json
import os
import pandas
import sh
import sys
import typing

def kgtk(arg1: typing.Union[str, pandas.DataFrame],
         arg2: typing.Optional[str] = None,
         df: typing.Optional[pandas.DataFrame] = None,
         auto_display_html: typing.Optional[bool] = None,
         auto_display_json: typing.Optional[bool] = None,
         auto_display_md: typing.Optional[bool] = None,
         unquote_column_names: typing.Optional[bool] = None,
         bash_command: typing.Optional[str] = None,
         kgtk_command: typing.Optional[str] = None,
         )->typing.Optional[pandas.DataFrame]:
    """This function simplifies using KGTK commands in a Jupyter Lab environment.
    Invocation
    ==========

    kgtk("pipeline")

        Execute the command pipeline.  The results are printed, displayed, or
    returned as a Pandas DataFrame.

    kgtk(df, "pipeline")

        The `df` in the call is a Pandas DataFrame, which is converted to KGTK
    format and passed to the pipeline as standard input. The results are
    printed, displayed, or returned as a Pandas DataFrame.

    Optional Parameters
    ======== ==========

    df=DF (default None)
    
        This is an alternate method for specifying an input DataFrame.

    auto_display_html=True/False (default True)

        This parameter controls the processing of HTML output.  See below.

    auto_display_json=True/False (default True)

        This parameter controls the processing of JSON output.  See below.

    auto_display_md=True/False (default False)

        This parameter controls the processing of MarkDown output.  See below.

    unquote_column_names=True/False (default True)

        Convert string column names to symbols.

    bash_command=CMD (default 'bash')

        This parameter specifies the name of the shell interpreter.  If the
    envar KGTK_BASH_COMMAND is present, it will supply the default value for
    the name of the shell interpreter.

    kgtk_command=CMD (default 'kgtk')

        This parameter specifies the kgtk shell command.  If the envar
    KGTK_KGTK_COMMAND is present, it will supply the default value for the
    name of the `kgtk` command.

        One use for this feature is to redefine the `kgtk` command to include
    `time` as a prefix, and/or to include common options.

    Standard Output Processing
    ======== ====== =========

    If the standard output of the pipeline is in HTML format (`--output-format HTML` or
    `kgtk("... /html")`), identified by starting with `<!DOCTYPE html>`, the
    output will be displayed with `display(HTML(output))` by default.
    However, if `kgtk(... auto_display_json=False)` or if the envar
    `KGTK_AUTO_DISPLAY_HTML` set to `false`, then the output will be printed.

    If the standard output of the pipeline is in JSON format (`--output-format JSON`),
    identified as starting with `[` or '{', the output will be displayed with
    `display(JSON(output))` by default.  However, if
    `kgtk(... auto_display_json=False)` or if the envar
    `KGTK_AUTO_DISPLAY_JSON` set to `false`, then the output will be printed.

    If the standard output of the pipeline is MarkDown format (typically by
    ending the pipeline in `... / md` or `... /table`, identified as starting
    with `|`, the output will be printed by default.  However, if
    `auto_display_md=True` is passed in the `kgtk(...)` call, or if the envar
    `KGTK_AUTO_DISPLAY_MD` is set to `true`, then the MarkDown will be
    displayed using `display(Markdown(output))`.

    If the standard output of the pipeline begins with "usage:", then it is
    treated as output from `--help` and printed.

    If the standard output starts with anything other than the cases listed
    above, then the output is assumed to be in KGTK format.  It is converted
    to a Pandas DataFrame, which is returned to the caller.

    Error Output Processing
    ===== ====== ==========

    If standard output was printed or displayed, then any error output will be printed
    immediately after it.

    If standard output was convertd to a DataFrame and returned, and
    subsequently displayed by the iPython shell, then any error output will be
    printed before the DataFrame is displayed.

    Environment Variables
    =========== =========

    This module directly uses the following environment variables:

    KGTK_AUTO_DISPLAY_HTML
    KGTK_AUTO_DISPLAY_JSON
    KGTK_AUTO_DISPLAY_MD
    KGTK_UNQUOTE_COLUMN_NAMES
    KGTK_BASH_COMMAND
    KGTK_KGTK_COMMAND

    """

    # Important prefixes to look for in standard output:
    MD_SIGIL: str = "|"
    JSON_SIGIL: str = "["
    JSONL_MAP_SIGIL: str = "{"
    HTML_SIGIL: str = "<!DOCTYPE html>"
    USAGE_SIGIL: str = "usage:" # Output from `kgtk --help` or `kgtk command --help`
    GRAPH_CACHE_SIGIL: str = "Graph Cache" # Output from `kgtk query --show-cache`

    # Set the defaults:
    if auto_display_html is None:
        auto_display_html = os.getenv("KGTK_AUTO_DISPLAY_HTML", "true").lower() in ["true", "yes", "y"]
    if auto_display_json is None:
        auto_display_json = os.getenv("KGTK_AUTO_DISPLAY_JSON", "true").lower() in ["true", "yes", "y"]
    if auto_display_md is None:
        auto_display_md = os.getenv("KGTK_AUTO_DISPLAY_MD", "false").lower() in ["true", "yes", "y"]
    if unquote_column_names is None:
        unquote_column_names = os.getenv("KGTK_UNQUOTE_COLUMN_NAMES", "true").lower() in ["true", "yes", "y"]

    # Why not os.getenv("KGTK_BASH_COMMAND", "bash")? Splitting it up makes
    # mypy happier.
    if bash_command is None:
        bash_command = os.getenv("KGTK_BASH_COMMAND")
    if bash_command is None:
        bash_command = "bash"

    if kgtk_command is None:
        kgtk_command = os.getenv("KGTK_KGTK_COMMAND")
    if kgtk_command is None:
        kgtk_command = "kgtk"

    # Figure out the input DataFrame and pipeline arguments:
    in_df: typing.Optional[pandas.DataFrame] = None
    pipeline: str
    if isinstance(arg1, str):
        if arg2 is not None:
            raise ValueError("kgtk(arg1, arg2): arg2 is not allowed when arg1 is a string")
        pipeline = arg1
    elif isinstance(arg1, pandas.DataFrame):
        if arg2 is None:
            raise ValueError("kgtk(arg1, arg2): arg2 is required when arg1 is a DataFrame")
        in_df = arg1
        pipeline = arg2

    if df is not None:
        if in_df is not None:
            raise ValueError("kgtk(): df= is not allowed when arg1 is a DataFrame")
        in_df = df

    if len(pipeline) == 0:
        raise ValueError("kgtk(...): the pipeline is empty")
    pipeline = kgtk_command + " " + ' '.join(pipeline.splitlines())

    # If we were supplied an input DataFrame, convert it to KGTK format.
    #
    # TODO: The conversion should optionally escape internal `|` characters as `\|`.
    in_tsv: typing.Optional[str] = None
    if in_df is not None:
        in_tsv = in_df.to_csv(sep='\t',
                              index=False,
                              quoting=csv.QUOTE_NONNUMERIC,
                              quotechar='"',
                              doublequote=False,
                              escapechar='\\',
                              )
        if unquote_column_names:
            # Pandas will have treated the column names as strings and quoted
            # them.  By convention, KGTK column names are symbols.  So, we will
            # remove double quotes from the outside of each column name.
            #
            # TODO: Handle the troublesome case of a double quote inside a column
            # name.
            header, body = in_tsv.split('\n', 1)
            column_names = header.split('\t')
            column_names = [x[1:-1] if x.startswith('"') else x for x in column_names ]
            header = "\t".join(column_names)
            in_tsv = header + "\n" + body

    # Execute the KGTK command pipeline:
    outbuf: StringIO = StringIO()
    errbuf: StringIO = StringIO()

    try:
        sh_bash = sh.Command(bash_command)
        p = sh_bash("-c", pipeline, _in=in_tsv, _out=outbuf, _err=errbuf, _bg=True)
        try:
            p.wait()

        except KeyboardInterrupt:
            print("kgtk: received KeyboardInterrupt", file=sys.stderr, flush=True)
            try:
                p.terminate()
            except Exception:
                pass

    except sh.ErrorReturnCode as e:
        # The pipeline returned an error.  stderr should hav ean error message.
        errmsg: str = errbuf.getvalue()
        if len(errmsg) > 0:
            print(errbuf.getvalue())
        else:
            print(str(e))
        return None

    output: str = outbuf.getvalue()

    # Decide what to do based on the start of the output:
    result: typing.Optional[pandas.DataFrame] = None
    if len(output) == 0:
        pass # No standard output
    
    elif output.startswith(MD_SIGIL):
        # Process Markdown output.
        if auto_display_md:
            display(Markdown(output))
        else:
            print(output)

    elif output.startswith(JSON_SIGIL) or output.startswith(JSONL_MAP_SIGIL):
        # Process JSON output.
        if auto_display_json:
            display(JSON(json.loads(output)))
        else:
            print(output)

    elif output[:len(HTML_SIGIL)].casefold() == HTML_SIGIL.casefold():
        # Process HTML output.
        if auto_display_html:
            display(HTML(output))
        else:
            print(output)

    elif output[:len(USAGE_SIGIL)].casefold() == USAGE_SIGIL.casefold():
        # Process --help output.
        print(output)

    elif output[:len(GRAPH_CACHE_SIGIL)].casefold() == GRAPH_CACHE_SIGIL.casefold():
        # Process `kgtk query --show-cache` output.
        print(output)

    else:
        # Assume that anything else is KGTK formatted output.  Convert it to a
        # pandas DataFrame and return it.
        #
        # TODO: Test this conversion with all KTK datatypes.  Language-qualified
        # strings are problematic.  Check what happens to quantites, date/times,
        # and locations.
        #
        # TODO: Remove the escape character from internal `|` characters?
        # If we do that, should we detect KGTK lists and complain?
        # `\|` -> `|`
        outbuf.seek(0)
        result = pandas.read_csv(outbuf,
                                 sep='\t',
                                 quotechar='"',
                                 doublequote=False,
                                 escapechar='\\',
                                 )

    outbuf.close()

    # Any error messages? If so, print the at the end.
    errout: str = errbuf.getvalue()
    if len(errout) > 0:
        print(errout)

    return result

def kypher(arg1: typing.Union[str, pandas.DataFrame],
           arg2: typing.Optional[str] = None,
           df: typing.Optional[pandas.DataFrame] = None,
           auto_display_html: typing.Optional[bool] = None,
           auto_display_json: typing.Optional[bool] = None,
           auto_display_md: typing.Optional[bool] = None,
           bash_command: typing.Optional[str] = None,
           kgtk_command: typing.Optional[str] = None,
           )->typing.Optional[pandas.DataFrame]:
    """This function simplifies using kypher (`kgtk query` in a Jupyter Lab environment.

    Invocation
    ==========

    kypher("pipeline")

        Execute the command pipeline with "query " prepended.  The results are
    printed, displayed, or returned as a Pandas DataFrame.

    kypher(df, "pipeline")

        The `df` in the call is a Pandas DataFrame, which is converted to KGTK
    format and passed to the pipeline as standard input. "query " will be
    prepended to the pipeline. The results are printed, displayed, or returned
    as a Pandas DataFrame.

    Environment Variables
    =========== =========

    KGTK_GRAPH_CACHE

        `kgtk query` and other KGTK commands use this environment variable to
    supply the location of the graph cache when `--graph cache` is not
    supplied on the command line.  In the JupyterLab environment
    KGTK_GRAPH_CACHE may be set with:

    %env KGTK_GRAPH_CACHE=/path/to/file.sqlite3.db


    Additional Details
    ========== =======

    See the documentation for the `kgtk(...)` for additional details

    """

    # Figure out the input DataFrame and pipeline arguments:
    in_df: typing.Optional[pandas.DataFrame] = None
    pipeline: str
    if isinstance(arg1, str):
        if arg2 is not None:
            raise ValueError("kypher(arg1, arg2): arg2 is not allowed when arg1 is a string")
        pipeline = arg1
    elif isinstance(arg1, pandas.DataFrame):
        if arg2 is None:
            raise ValueError("kypher(arg1, arg2): arg2 is required when arg1 is a DataFrame")
        in_df = arg1
        pipeline = arg2

    if df is not None:
        if in_df is not None:
            raise ValueError("kypher(): df= is not allowed when arg1 is a DataFrame")
        in_df = df

    pipeline = "query " + pipeline

    return kgtk(pipeline,
                df=in_df,
                auto_display_html=auto_display_html,
                auto_display_json=auto_display_json,
                auto_display_md=auto_display_md,
                bash_command=bash_command,
                kgtk_command=kgtk_command,
                )



import shutil
import tempfile
import unittest
from kgtk.cli_entry import cli_entry


class TestVisualizeGraph(unittest.TestCase):
    def setUp(self) -> None:
        self.example_file = 'data/visualize_force_graph_example2.tsv'
        self.node_file = 'data/visualize_force_graph_example2_node.tsv'
        self.node_file_blank_labels = 'data/visualize_graph_node_example_blank_labels.tsv'
        self.ground_truth_default = 'data/visualize_graph_example_1_no_node_default.html'
        self.ground_truth_color_node = 'data/visualize_graph_example_color_by_node_column.html'
        self.ground_truth_color_node_log = 'data/visualize_graph_example_color_by_node_column_log.html'
        self.ground_truth_color_node_missing = 'data/visualize_graph_example_color_by_node_column_log_missing.html'
        self.ground_truth_color_node_hex = 'data/visualize_graph_example_color_by_node_column_hex.html'
        self.ground_truth_color_edge = 'data/visualize_graph_example_color_edge.html'
        self.ground_truth_color_edge_numbers = 'data/visualize_graph_example_color_edge_numbers.html'
        self.ground_truth_color_edge_strings = 'data/visualize_graph_example_color_edge_strings.html'
        self.ground_truth_node_size = 'data/visualize_graph_example_node_size.html'
        self.ground_truth_edge_width = 'data/visualize_graph_example_edge_width.html'
        self.ground_truth_node_text = 'data/visualize_graph_example_node_text.html'
        self.ground_truth_node_text_number_colors = 'data/visualize_graph_example_node_text_number_colors.html'
        self.ground_truth_edge_text = 'data/visualize_graph_example_edge_text.html'
        self.ground_truth_node_edge_text = 'data/visualize_graph_example_node_edge_text.html'
        self.ground_truth_node_text_blank_labels = 'data/visualize_graph_example_node_text_blank_labels.html'
        self.ground_truth_color_node_custom_interpolator = 'data/visualize_graph_example_color_by_node_column_interpolator.html'
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_default_no_node_file(self):
        output = f'{self.temp_dir}/test_1.html'
        cli_entry("kgtk", "--debug", "visualize-graph", "-i", self.example_file, "-o", output)

        f = open(self.ground_truth_default)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_node_column(self):
        output = f'{self.temp_dir}/test_2.html'
        cli_entry("kgtk", "--debug", "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--node-color-column", "is_country",
                  "--node-color-style", "d3.schemeCategory10",
                  "--node-color-numbers", "as-is",
                  "--node-file", f'{self.node_file}'
                  )
        f = open(self.ground_truth_color_node)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_node_column_log(self):
        output = f'{self.temp_dir}/test_3.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--node-color-column", "degree",
                  "--node-color-style", "d3.schemePastel1",
                  "--node-color-numbers", "log",
                  "--node-file", f'{self.node_file}'
                  )

        f = open(self.ground_truth_color_node_log)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_node_column_missing(self):
        output = f'{self.temp_dir}/test_4.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--node-color-column", "type_missing",
                  "--node-color-style", "d3.schemeSet3",
                  "--node-file", f'{self.node_file}'
                  )

        f = open(self.ground_truth_color_node_missing)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_node_column_hex(self):
        output = f'{self.temp_dir}/test_5.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--node-color-column", "hex_color",
                  "--node-color-hex",
                  "--node-file", f'{self.node_file}'
                  )

        f = open(self.ground_truth_color_node_hex)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_edge_column(self):
        output = f'{self.temp_dir}/test_6.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--edge-color-column", "hex_color",
                  "--edge-color-hex"
                  )

        f = open(self.ground_truth_color_edge)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_node_size(self):
        output = f'{self.temp_dir}/test_7.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file}',
                  "-o", f'{output}',
                  "--node-size-column", "population",
                  "--node-size-minimum", "2.0",
                  "--node-size-maximum", "6.0",
                  "--node-size-default", "4.0",
                  "--node-color-column", "hex_color",
                  "--node-color-hex",
                  "--node-size-scale", "log"
                  )
        f = open(self.ground_truth_node_size)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_edge_width(self):
        output = f'{self.temp_dir}/test_8.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file}',
                  "-o", f'{output}',
                  "--edge-width-column", "weight",
                  "--edge-width-minimum", "2.0",
                  "--edge-width-maximum", "5.0",
                  "--edge-width-default", "2.0",
                  "--edge-width-scale", "log",
                  "--node-color-column", "degree",
                  "--node-color-style", "d3.schemeDark2",
                  "--node-color-numbers", "linear"
                  )

        f = open(self.ground_truth_edge_width)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_node_text(self):
        output = f'{self.temp_dir}/test_9.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file}',
                  "-o", f'{output}',
                  "--node-color-column", "hex_color",
                  "--node-color-hex",
                  "--show-text", "above"
                  )

        f = open(self.ground_truth_node_text)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_edge_text(self):
        output = f'{self.temp_dir}/test_10.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--show-edge-label",
                  "--edge-color-column", "hex_color",
                  "--edge-color-hex"
                  )

        f = open(self.ground_truth_edge_text)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_node_edge_text(self):
        output = f'{self.temp_dir}/test_11.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file}',
                  "-o", f'{output}',
                  "--node-color-column", "hex_color",
                  "--node-color-hex",
                  "--show-text", "above",
                  "--show-edge-label",
                  "--edge-color-hex",
                  "--edge-color-column", "hex_color"
                  )

        f = open(self.ground_truth_node_edge_text)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_node_text_blank_labels(self):
        output = f'{self.temp_dir}/test_12.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file_blank_labels}',
                  "-o", f'{output}',
                  "--node-color-column", "hex_color",
                  "--node-color-hex",
                  "--show-text", "above",
                  "--show-blank-labels"
                  )

        f = open(self.ground_truth_node_text_blank_labels)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_node_column_custom_interpolator(self):
        output = f'{self.temp_dir}/test_13.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--node-color-column", "ordinal",
                  "--node-color-numbers", "as-is",
                  "--node-file", f'{self.node_file}',
                  "--node-color-style", "d3.interpolateGreens"
                  )

        f = open(self.ground_truth_color_node_custom_interpolator)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_node_text_numbers_colors(self):
        output = f'{self.temp_dir}/test_14.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "--node-file", f'{self.node_file}',
                  "-o", f'{output}',
                  "--node-color-column", "ordinal",
                  "--node-color-numbers", "as-is",
                  "--node-color-style", "d3.interpolateGreens",
                  "--show-text", "above"
                  )

        f = open(self.ground_truth_node_text_number_colors)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_edge_column_color_numbers(self):
        output = f'{self.temp_dir}/test_15.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--edge-color-column", "ordinal",
                  "--edge-color-numbers", "as-is",
                  "--edge-color-style", "d3.interpolateReds"
                  )

        f = open(self.ground_truth_color_edge_numbers)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)

    def test_color_by_edge_column_color_strings(self):
        output = f'{self.temp_dir}/test_16.html'
        cli_entry("kgtk", "--debug",
                  "visualize-graph",
                  "-i", self.example_file,
                  "-o", f'{output}',
                  "--edge-color-column", "label",
                  "--edge-color-style", "d3.schemeDark2"
                  )

        f = open(self.ground_truth_color_edge_strings)
        f1 = set(f.readlines())
        f.close()
        with open(output) as f2:
            for line in f2:
                self.assertTrue(line in f1)


if __name__ == '__main__':
    unittest.main()



import graph_tool as gtmain
import graph_tool.all as gtall
import graph_tool.topology as gttop


def get_nodes_with_degree(g, deg_from, deg_to):
    u = gtmain.GraphView(g, vfilt=lambda v: v.in_degree()+v.out_degree() in range(deg_from,deg_to))
    return u

def get_nodes_by_node_prop(g, p, v):
    return gtall.find_vertex(g, prop=g.properties[('v', p)], match=v)

def get_edges_by_edge_prop(g, p, v):
    return gtall.find_edge(g, prop=g.properties[('e', p)], match=v)

def get_neighbors(g, node_id, direction):
    the_node=get_nodes_by_node_prop(g, '_graphml_vertex_id', node_id)[0]
    if direction=='out':
        return set(the_node.out_neighbors())
    elif direction=='in':
        return set(the_node.in_neighbors())
    else: # total
        return set(the_node.out_neighbors()) | set(the_node.in_neighbors())

def compute_transitive_closure(g):
    return gttop.transitive_closure(g)



from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles


def parser():
    return {
        'help': 'Convert embeddings format from KGTK edge file to word2vec or Google Projector format, or vice versa',
        'description': 'Converts KGTK edge embeddings file to word2vec or Google Projector format.'
                       'Takes an optional node file for Google Project format to create a metadata file. '
                       'Processes only top 10,000 rows from the edge file for Google Projector as it only'
                       ' accepts 10,000 rows.\n'
                       'Additional options are shown in expert help.\n'
                       'kgtk --expert convert-embeddings-format --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str) -> str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="KGTK input files",
                          dest="input_file",
                          options=["-i", "--input-file"],
                          positional=True,
                          allow_list=False)

    parser.add_input_file(who="The KGTK node file for creating Google Projector Metadata. All the columns in the node "
                              "file will be added to the metadata file by default. You can customise this with the "
                              "--metadata-columns option.",
                          options=["--node-file"],
                          dest="node_file",
                          metavar="NODE_FILE",
                          optional=True,
                          allow_stdin=False)

    parser.add_output_file()

    parser.add_output_file(who="The output metadata file for Google Projector. If --output-format == gprojector and"
                               "--metadata-file is not specified, a file named "
                               "`kgtk_embeddings_gprojector_metadata.tsv` will be created in USER_HOME",
                           options=["--metadata-file"],
                           dest="metadata_file",
                           metavar="METADATA_FILE",
                           optional=True,
                           allow_stdout=False,
                           default_stdout=False)

    parser.add_argument("--input-property",
                        dest="input_property",
                        help="The property name for embeddings in the input KGTK edge file. (default=embeddings).",
                        default='embeddings',
                        type=str)

    parser.add_argument("--output-format",
                        dest="output_format",
                        help="The desired output file format: word2vec|gprojector (default=word2vec)",
                        default="word2vec",
                        type=str)

    parser.add_argument("--metadata-columns",
                        dest="metadata_columns",
                        help="A comma separated string of columns names in the input file or the --node-file to be "
                             "used for creating the metadata file for Google projector. "
                             "Only to be used when --output-format == 'gprojector'. If --node-file is specified, "
                             "the command will look for --metadata-columns in the --node-file, otherwise input file."
                             "The command will throw an error if the columns specified are not in either of the files.",
                        default=None,
                        type=str)

    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        node_file: KGTKFiles,
        output_format: typing.Optional[str] = 'word2vec',
        input_property: typing.Optional[str] = 'embeddings',
        metadata_file: KGTKFiles = None,
        metadata_columns: str = None,
        errors_to_stderr: bool = True,
        **kwargs  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ) -> int:
    # import modules locally
    from kgtk.exceptions import KGTKException
    from kgtk.utils.convert_embeddings_format import ConvertEmbeddingsFormat

    input_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)
    node_file_path: Path = KGTKArgumentParser.get_input_file(node_file) if node_file is not None else None
    metadata_file_path: Path = KGTKArgumentParser.get_input_file(metadata_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stderr if errors_to_stderr else sys.stdout

    # Build the option structures.
    kwargs['use_graph_cache_envar'] = False

    try:
        cef: ConvertEmbeddingsFormat = ConvertEmbeddingsFormat(input_file=input_file_path,
                                                               output_file=output_file_path,
                                                               node_file=node_file_path,
                                                               output_format=output_format,
                                                               input_property=input_property,
                                                               error_file=error_file,
                                                               output_metadata_file=metadata_file_path,
                                                               metadata_columns=metadata_columns,
                                                               **kwargs
                                                               )

        cef.process()
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except KGTKException as e:
        raise KGTKException(e)
    except Exception as e:
        raise KGTKException(str(e))
